@@ -21,17 +21,20 @@ import functools
 import os
 import threading
 
+from tensorflow.compiler.xla.tsl.protobuf import coordination_config_pb2
 from tensorflow.python.distribute import cross_device_ops as cross_device_ops_lib
 from tensorflow.python.distribute import device_util
 from tensorflow.python.distribute import distribute_lib
 from tensorflow.python.distribute import input_lib
 from tensorflow.python.distribute import input_util
+from tensorflow.python.distribute import load_context
 from tensorflow.python.distribute import mirrored_run
 from tensorflow.python.distribute import multi_worker_util
 from tensorflow.python.distribute import parameter_server_strategy
 from tensorflow.python.distribute import ps_values
 from tensorflow.python.distribute import sharded_variable
 from tensorflow.python.distribute import values
+from tensorflow.python.distribute.cluster_resolver import cluster_resolver as base_cluster_resolver
 from tensorflow.python.distribute.coordinator import cluster_coordinator
 from tensorflow.python.eager import context
 from tensorflow.python.eager import remote
@@ -45,11 +48,9 @@ from tensorflow.python.ops import variable_scope as vs
 from tensorflow.python.platform import tf_logging as logging
 from tensorflow.python.trackable import base as trackable
 from tensorflow.python.training import server_lib
-from tensorflow.python.util import keras_deps
 from tensorflow.python.util import nest
 from tensorflow.python.util import tf_inspect
 from tensorflow.python.util.tf_export import tf_export
-from tensorflow.tsl.protobuf import coordination_config_pb2
 
 
 ALLOWED_TASK_TYPES = ("chief", "worker", "ps")
@@ -58,11 +59,21 @@ ALLOWED_TASK_TYPES = ("chief", "worker", "ps")
 # is used. Refer to the discussion in b/249134783 for more.
 _HEARTBEAT_TIMEOUT_SECS = 5
 
+# Set the number of retries during initial connection for fault tolerance.
+# Retries follow an exponential backoff waiting period as defined in the
+# runtime, with min value 1ms, max value 10s, and exponent 1.3. So 50 retries
+# enables ~3 minutes of retrying. In general, this means the first 35 retries
+# consist of 42 seconds of backoff waiting, and each subsequent retry waits for
+# 10 seconds. So to enable 30 minutes of retrying, we would want
+# 35 + (30 * 60 - 42) // 10 = 210 retries.
+_SET_SERVER_DEF_RETRIES = 50
+
 
 @tf_export(
     "distribute.experimental.ParameterServerStrategy",
     "distribute.ParameterServerStrategy",
-    v1=[])
+    v1=[],
+)
 class ParameterServerStrategyV2(distribute_lib.Strategy):
   """An multi-worker tf.distribute strategy with parameter servers.
 
@@ -376,7 +387,7 @@ class ParameterServerStrategyV2(distribute_lib.Strategy):
 
   When a partitioned variable is saved to a `SavedModel`, it will be saved as if
   it is one single variable. This improves serving efficiency by eliminating
-  a number of Ops that handle the partiton aspects.
+  a number of Ops that handle the partition aspects.
 
   Known limitations of variable partitioning:
 
@@ -424,7 +435,7 @@ class ParameterServerStrategyV2(distribute_lib.Strategy):
   """
 
   # pyformat: disable
-  def __init__(self, cluster_resolver, variable_partitioner=None):
+  def __init__(self, cluster_resolver: base_cluster_resolver.ClusterResolver, variable_partitioner: sharded_variable.Partitioner = None):
     """Initializes the TF2 parameter server strategy.
 
     This initializes the `tf.distribute.experimental.ParameterServerStrategy`
@@ -488,7 +499,7 @@ class ParameterServerStrategyV2(distribute_lib.Strategy):
     # Used to check if isinstance() without having to import this module
     self._is_parameter_server_strategy_v2 = True
 
-  def _configure_coordination_service(self, cluster_spec):
+  def _configure_coordination_service(self, cluster_spec: base_cluster_resolver.ClusterSpec):
     if context.context().coordination_service is None:
       coordinated_jobs = ["worker", "ps"]
       coordinated_job_config = []
@@ -505,7 +516,7 @@ class ParameterServerStrategyV2(distribute_lib.Strategy):
           heartbeat_timeout_in_ms=_HEARTBEAT_TIMEOUT_SECS * 1000,
           allow_new_incarnation_to_reconnect=True)
 
-  def _connect_to_cluster(self, coordinator_name):
+  def _connect_to_cluster(self, coordinator_name: str):
     if coordinator_name in ["worker", "ps"]:
       raise ValueError("coordinator name should not be 'worker' or 'ps'.")
     cluster_spec = self._cluster_resolver.cluster_spec()
@@ -545,7 +556,13 @@ class ParameterServerStrategyV2(distribute_lib.Strategy):
     distribute_lib.distribution_strategy_replica_gauge.get_cell(
         "ps_strategy_num_ps").set(self._num_ps)
 
-  def _verify_args_and_config(self, cluster_resolver):
+    # Explicitly connect to the cluster here. Enable retries in case of worker
+    # preemptions during connection.
+    context.set_server_def_retries(_SET_SERVER_DEF_RETRIES)
+    # Perform connection by initializing context.
+    context.ensure_initialized()
+
+  def _verify_args_and_config(self, cluster_resolver: base_cluster_resolver.ClusterResolver):
     if not cluster_resolver.cluster_spec():
       raise ValueError("Cluster spec must be non-empty in "
                        "`tf.distribute.cluster_resolver.ClusterResolver`.")
@@ -563,19 +580,25 @@ class ParameterServerStrategyV2(distribute_lib.Strategy):
 
 
 class ParameterServerStrategyV2Extended(
-    parameter_server_strategy.ParameterServerStrategyExtended):
+    parameter_server_strategy.ParameterServerStrategyExtended
+):
   """Extended class for ParameterServerStrategyV2.
 
   Please see `tf.distribute.StrategyExtended` doc for more information.
   """
 
-  def __init__(self, container_strategy, cluster_resolver,
-               variable_partitioner):
+  def __init__(
+      self,
+      container_strategy,
+      cluster_resolver: base_cluster_resolver.ClusterResolver,
+      variable_partitioner,
+  ):
     """Initialization of ParameterServerStrategyV2Extended."""
     super(ParameterServerStrategyV2Extended, self).__init__(container_strategy)
     self._num_ps = len(cluster_resolver.cluster_spec().as_dict().get("ps", []))
-    self._num_workers = len(cluster_resolver.cluster_spec().as_dict().get(
-        "worker", []))
+    self._num_workers = len(
+        cluster_resolver.cluster_spec().as_dict().get("worker", [])
+    )
     self._variable_count = 0
 
     self._variable_partitioner = variable_partitioner
@@ -585,14 +608,16 @@ class ParameterServerStrategyV2Extended(
     self._being_scheduled = False
     self._set_num_gpus()
     distribute_lib.distribution_strategy_replica_gauge.get_cell(
-        "num_gpus_per_worker").set(self._num_gpus_per_worker)
+        "num_gpus_per_worker"
+    ).set(self._num_gpus_per_worker)
 
     # Don't canonicalize the devices here since this code is executed on Chief,
     # but we want the reduce evaluation to be done on each worker. Placer will
     # automatically choose the right device based on current context.
     # TODO(ishark): Use select_cross_device_ops instead.
     self._cross_device_ops = cross_device_ops_lib.ReductionToOneDevice(
-        reduce_to_device="/device:CPU:0")
+        reduce_to_device="/device:CPU:0"
+    )
     self._cross_device_ops._canonicalize_devices = False  # pylint: disable=protected-access
     self._allow_run_without_coordinator = False
     self._coordinator_creation_lock = threading.Lock()
@@ -628,17 +653,24 @@ class ParameterServerStrategyV2Extended(
       # Create and wrap the variable.
       v = next_creator(**kwargs)
       wrapped_v = ps_values.CachingVariable(v)
-      wrapped = ps_values.AggregatingVariable(self._container_strategy(),
-                                              wrapped_v, aggregation)
+      wrapped = ps_values.AggregatingVariable(
+          self._container_strategy(), wrapped_v, aggregation
+      )
       return wrapped
 
     if self._num_replicas_in_sync > 1:
-      if aggregation not in (vs.VariableAggregation.NONE,
-                             vs.VariableAggregation.SUM,
-                             vs.VariableAggregation.MEAN,
-                             vs.VariableAggregation.ONLY_FIRST_REPLICA):
-        raise ValueError("Invalid variable aggregation mode: " + aggregation +
-                         " for variable: " + kwargs["name"])
+      if aggregation not in (
+          vs.VariableAggregation.NONE,
+          vs.VariableAggregation.SUM,
+          vs.VariableAggregation.MEAN,
+          vs.VariableAggregation.ONLY_FIRST_REPLICA,
+      ):
+        raise ValueError(
+            "Invalid variable aggregation mode: "
+            + aggregation
+            + " for variable: "
+            + kwargs["name"]
+        )
       return var_creator
     else:
 
@@ -648,6 +680,12 @@ class ParameterServerStrategyV2Extended(
 
       return variable_creator_single_replica
 
+  def _create_per_worker_variable(self, next_creator, **kwargs):
+    """Create an unsynced, unaggregated variable on each worker."""
+    return ps_values.PerWorkerVariable(
+        self._container_strategy(), next_creator, **kwargs
+    )
+
   def _create_variable(self, next_creator, **kwargs):
     """Implements StrategyExtendedV2._create_variable.
 
@@ -667,6 +705,9 @@ class ParameterServerStrategyV2Extended(
     Returns:
       A `Variable` or `ShardedVariable`.
     """
+    if kwargs.pop("per_worker_variable", False):
+      logging.info("Creating per worker variable")
+      return self._create_per_worker_variable(next_creator, **kwargs)
 
     var_creator = self._create_var_creator(next_creator, **kwargs)
     if "colocate_with" in kwargs:  # Never partition colocated_with variables.
@@ -678,7 +719,10 @@ class ParameterServerStrategyV2Extended(
           var = var_creator(**kwargs)
           logging.debug(
               "Creating variable (name:%s, shape:%r) that colocates with %s",
-              var.name, var.shape, kwargs["colocate_with"].name)
+              var.name,
+              var.shape,
+              kwargs["colocate_with"].name,
+          )
           return var
 
     if self._variable_partitioner is None:
@@ -693,20 +737,24 @@ class ParameterServerStrategyV2Extended(
       v = next_creator(**kwargs)
       if not isinstance(v, resource_variable_ops.UninitializedVariable):
         raise ValueError(
-            "It looks like you are using `ParameterServerStrategy` with a "
-            "`variable_partitioner`, and trying to create a variable without "
-            "specifying `initial_value`. This is not allowed. Please specify the "
-            "`initial_value`.")
+            "It looks like you are using `ParameterServerStrategy` with a"
+            " `variable_partitioner`, and trying to create a variable without"
+            " specifying `initial_value`. This is not allowed. Please specify"
+            " the `initial_value`."
+        )
       elif shape is None or dtype is None:
         raise ValueError(
             "It looks like you are trying to load a `SavedModel` using "
             "`tf.saved_model.load` within a `ParameterServerStrategy` scope, "
-            "but the `SavedModel` is missing shape or dtype information.")
+            "but the `SavedModel` is missing shape or dtype information."
+        )
       else:
+
         def initializer(shape, dtype, **kwargs):
           if "partition_shape" in kwargs:
             shape = kwargs["partition_shape"]
           return array_ops.zeros(shape, dtype)
+
         initial_value = functools.partial(initializer, shape=shape, dtype=dtype)
 
     # Two cases where initial_value can be a callable:
@@ -732,11 +780,15 @@ class ParameterServerStrategyV2Extended(
       return self._create_variable_round_robin(var_creator, **kwargs)
 
     num_partitions = self._variable_partitioner(shape=shape, dtype=dtype)
-    if not num_partitions or num_partitions[0] == 0 or any(
-        v != 1 for v in num_partitions[1:]):
+    if (
+        not num_partitions
+        or num_partitions[0] == 0
+        or any(v != 1 for v in num_partitions[1:])
+    ):
       raise ValueError(
           "variable_partitioner must return a list/tuple whose elements are 1"
-          " besides the first element (non-zero), got: %r" % num_partitions)
+          " besides the first element (non-zero), got: %r" % num_partitions
+      )
 
     if num_partitions[0] == 1:  # no partition
       return self._create_variable_round_robin(var_creator, **kwargs)
@@ -759,19 +811,25 @@ class ParameterServerStrategyV2Extended(
     def init_shard_fn(shard_index):
       if not init_from_fn:
         logging.log_if(
-            logging.WARN, _INEFFICIENT_INIT_WARNING % name, shard_index == 0 and
-            shape.num_elements() > _LARGE_VARIABLE_NUM_ELEMENTS)
-        return initial_value[offsets[shard_index]:offsets[shard_index + 1]]
-      partition_shape = (offsets[shard_index + 1] -
-                         offsets[shard_index],) + shape[1:]
+            logging.WARN,
+            _INEFFICIENT_INIT_WARNING % name,
+            shard_index == 0
+            and shape.num_elements() > _LARGE_VARIABLE_NUM_ELEMENTS,
+        )
+        return initial_value[offsets[shard_index] : offsets[shard_index + 1]]
+      partition_shape = (
+          offsets[shard_index + 1] - offsets[shard_index],
+      ) + shape[1:]
       partition_offset = (offsets[shard_index],) + (0,) * len(shape[1:])
       arg_spec = tf_inspect.getfullargspec(initial_value)
-      if ("shard_info" not in arg_spec.args and
-          "shard_info" not in arg_spec.kwonlyargs):
+      if (
+          "shard_info" not in arg_spec.args
+          and "shard_info" not in arg_spec.kwonlyargs
+      ):
         try:
           value = initial_value(
-              partition_shape=partition_shape,
-              partition_offset=partition_offset)
+              partition_shape=partition_shape, partition_offset=partition_offset
+          )
         except (TypeError, ValueError):
           # TypeError: Initializer doesn't accept kwargs
           # ValueError: Initializer doesn't accept partition kwargs
@@ -785,16 +843,20 @@ class ParameterServerStrategyV2Extended(
           # Initializer doesn't support partition: value is the full value
           # and needs to be sliced to get the partition value.
           logging.log_if(
-              logging.WARN, _INEFFICIENT_INIT_WARNING % name,
-              shard_index == 0 and
-              shape.num_elements() > _LARGE_VARIABLE_NUM_ELEMENTS)
-          return value[offsets[shard_index]:offsets[shard_index + 1]]
+              logging.WARN,
+              _INEFFICIENT_INIT_WARNING % name,
+              shard_index == 0
+              and shape.num_elements() > _LARGE_VARIABLE_NUM_ELEMENTS,
+          )
+          return value[offsets[shard_index] : offsets[shard_index + 1]]
       else:
         # For compatibility with `CheckpointInitialValueCallable`.
         return initial_value(
             shard_info=trackable.ShardInfo(
                 shape=tensor_shape.as_shape(partition_shape),
-                offset=partition_offset))
+                offset=partition_offset,
+            )
+        )
 
     var_list = []
     for i in range(num_partitions):
@@ -813,13 +875,22 @@ class ParameterServerStrategyV2Extended(
     with ops.colocate_with(None, ignore_existing=True):
       # Explicitly set CPU:0 device for PS in case create variable is called
       # inside replica_fn and worker has with GPU:0 scope.
-      with ops.device("/job:ps/task:%d/device:CPU:0" %
-                      (self._variable_count % self._num_ps)):
+      with ops.device(
+          "/job:ps/task:%d/device:CPU:0" % (self._variable_count % self._num_ps)
+      ):
         var = next_creator(**kwargs)
-        logging.debug(
+        log_method = (
+            logging.info
+            if os.getenv("TF_PSS_VERBOSE_VARIABLE_PLACEMENT")
+            else logging.debug
+        )
+        log_method(
             "Creating variable (name:%s, shape:%r) on "
-            "/job:ps/task:%d/device:CPU:0", var.name, var.shape,
-            (self._variable_count % self._num_ps))
+            "/job:ps/task:%d/device:CPU:0",
+            var.name,
+            var.shape,
+            (self._variable_count % self._num_ps),
+        )
         self._variable_count += 1
         return var
 
@@ -828,7 +899,8 @@ class ParameterServerStrategyV2Extended(
     with self._coordinator_creation_lock:
       if not self._container_strategy()._cluster_coordinator:  # pylint: disable=protected-access
         cluster_coordinator.ClusterCoordinator(
-            strategy=self._container_strategy())
+            strategy=self._container_strategy()
+        )
 
     # TODO(wxinyi): We should warn the user of the inefficiency of creating
     # `StaticHashTable` inside a `@tf.function`-wrapped `dataset_fn` to be
@@ -841,32 +913,39 @@ class ParameterServerStrategyV2Extended(
     # the coordinator which incurs worker-coordinator communication overhead.
 
     def lookup_creator(next_creator, *args, **kwargs):
-      if keras_deps.get_load_context_function()():
-        return (ps_values.RestoredDistributedTable(
-            self._container_strategy(), lambda: next_creator(*args, **kwargs)))  # pylint: disable=protected-access
+      if load_context.in_load_context():
+        return ps_values.RestoredDistributedTable(
+            self._container_strategy(), lambda: next_creator(*args, **kwargs)
+        )  # pylint: disable=protected-access
       else:
-        return ps_values.DistributedTable(self._container_strategy(),
-                                          lambda: next_creator(*args, **kwargs))  # pylint: disable=protected-access
+        return ps_values.DistributedTable(
+            self._container_strategy(), lambda: next_creator(*args, **kwargs)
+        )  # pylint: disable=protected-access
 
     def restored_lookup_creator(next_creator, *args, **kwargs):
-      return (ps_values.RestoredDistributedTable(
-          self._container_strategy(), lambda: next_creator(*args, **kwargs)))  # pylint: disable=protected-access
+      return ps_values.RestoredDistributedTable(
+          self._container_strategy(), lambda: next_creator(*args, **kwargs)
+      )  # pylint: disable=protected-access
 
     return [
         ops.resource_creator_scope("StaticHashTable", lookup_creator),
-        ops.resource_creator_scope("RestoredStaticHashTable",
-                                   restored_lookup_creator)
+        ops.resource_creator_scope(
+            "RestoredStaticHashTable", restored_lookup_creator
+        ),
     ]
 
   def _assert_used_with_cluster_coordinator(self):
-    if (not self._used_with_coordinator and
-        not self._allow_run_without_coordinator):
+    if (
+        not self._used_with_coordinator
+        and not self._allow_run_without_coordinator
+    ):
       raise NotImplementedError(
           "`tf.distribute.experimental.ParameterServerStrategy` must be used "
           "with `tf.distribute.experimental.coordinator.ClusterCoordinator` in "
           "a custom training loop. If you are using `Model.fit`, please supply "
           "a dataset function directly to a "
-          "`tf.keras.utils.experimental.DatasetCreator` instead.")
+          "`tf.keras.utils.experimental.DatasetCreator` instead."
+      )
 
   def _assert_being_scheduled_by_cluster_coordinator(self):
     if not self._being_scheduled and not self._allow_run_without_coordinator:
@@ -877,14 +956,16 @@ class ParameterServerStrategyV2Extended(
           "coordinator, which can be slow. To properly dispatch functions to "
           "run on workers, methods like `run` or `reduce` should be used "
           "within a function passed to `tf.distribute.experimental.coordinator."
-          "ClusterCoordinator.schedule`.")
+          "ClusterCoordinator.schedule`."
+      )
 
   # options is not used right now. But we may want to support options while
   # creating InputWorkers in future, similar to MirroredStrategy.
   def _input_workers_with_options(self, options=None):
     input_workers_devices = (("/device:CPU:0", self.worker_devices),)
     return input_lib.InputWorkers(
-        input_workers_devices, canonicalize_devices=False)
+        input_workers_devices, canonicalize_devices=False
+    )
 
   def _experimental_distribute_dataset(self, dataset, options):
     input_workers_devices = self._input_workers_with_options()
@@ -898,7 +979,8 @@ class ParameterServerStrategyV2Extended(
         self._container_strategy(),
         num_replicas_in_sync=self._num_replicas_in_sync,
         options=options,
-        build=ops.inside_function())  # will be built by ClusterCoordinator
+        build=ops.inside_function(),
+    )  # will be built by ClusterCoordinator
 
   def _distribute_datasets_from_function(self, dataset_fn, options):
     # There is no synchronization beyond a worker and thus, the number of
@@ -909,7 +991,8 @@ class ParameterServerStrategyV2Extended(
     input_context = distribute_lib.InputContext(
         num_input_pipelines=num_input_pipelines_in_sync,
         input_pipeline_id=input_pipeline_id_in_sync,
-        num_replicas_in_sync=self._num_replicas_in_sync)
+        num_replicas_in_sync=self._num_replicas_in_sync,
+    )
 
     # If this DistributedDatasetFromFunction is created outside
     # ClusterCoordinator, i,e, outside a tf.function, we don't build its
@@ -917,10 +1000,12 @@ class ParameterServerStrategyV2Extended(
     # ClusterCoordinator.create_per_worker_dataset.
     return input_util.get_distributed_datasets_from_function(
         dataset_fn,
-        self._input_workers_with_options(options), [input_context],
+        self._input_workers_with_options(options),
+        [input_context],
         self._container_strategy(),
         options=options,
-        build=ops.inside_function())  # will be built by ClusterCoordinator
+        build=ops.inside_function(),
+    )  # will be built by ClusterCoordinator
 
   @property
   def worker_devices(self):
@@ -934,15 +1019,17 @@ class ParameterServerStrategyV2Extended(
   def _call_for_each_replica(self, fn, args, kwargs):
     self._assert_being_scheduled_by_cluster_coordinator()
 
-    return mirrored_run.call_for_each_replica(self._container_strategy(), fn,
-                                              args, kwargs)
+    return mirrored_run.call_for_each_replica(
+        self._container_strategy(), fn, args, kwargs
+    )
 
   def _reduce(self, reduce_op, value):
     self._assert_being_scheduled_by_cluster_coordinator()
     dst = device_util.current() or self._default_device or "/device:CPU:0"
     destinations = device_util.canonicalize_without_job_and_task(dst)
     result = self._local_results(
-        self.reduce_to(reduce_op, value, destinations))[0]
+        self.reduce_to(reduce_op, value, destinations)
+    )[0]
     return result
 
   def _reduce_to(self, reduce_op, value, destinations, options):
@@ -951,7 +1038,8 @@ class ParameterServerStrategyV2Extended(
     def get_values(x):
       if isinstance(x, values.DistributedValues):
         return self._cross_device_ops.reduce(
-            reduce_op, x, destinations=destinations)  # pylint: disable=protected-access
+            reduce_op, x, destinations=destinations
+        )  # pylint: disable=protected-access
       return x
 
     return nest.map_structure(get_values, value)
@@ -966,6 +1054,7 @@ _INEFFICIENT_INIT_WARNING = (
     "footprint, explicitly specify `dtype` and `shape` when creating "
     "variables, and use `tf.initializers` to initialize the variable. "
     "Note that some initializers (e.g., orthogonal) don't support "
-    "memory-efficient initialization and there is not much you can do here.")
+    "memory-efficient initialization and there is not much you can do here."
+)
 
 _LARGE_VARIABLE_NUM_ELEMENTS = 1e9
